package org.processmining.glueminer.algorithms;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.glueminer.models.GeneratedPetrinet;
import org.processmining.glueminer.models.PlacesMemorizer;
import org.processmining.glueminer.models.PluginContextFactory;
import org.processmining.glueminer.parameters.GeneratePetrinetParameters;
import org.processmining.glueminer.parameters.LogGeneratorParameters;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.semantics.petrinet.Marking;

public class GeneratePetrinet {
	PluginContext context;
	int numStartActivities;
	int numEndActivities;
	int noIntermediateActivities;
	int noOfWantedPlaces;
	int maxNoIterations;
	int maxSourceActivitiesNumber;
	int maxTargetActivitiesNumber;
	
	Petrinet net;
	Marking initialMarking;
	Marking finalMarking;
	Place sourcePlace;
	Place sinkPlace;
	//Place intStartActivities;
	//Place intEndActivities;
	
	List<String> sourceActivitiesString;
	List<String> targetActivitiesString;
	List<String> intermediateActivitiesString;
	Map<String, Transition> transitionCorrespondence;
	
	PlacesMemorizer placesMemorizer;
	
	public GeneratePetrinet(PluginContext context, GeneratePetrinetParameters parameters) {
		this.context = context;
		
		this.numStartActivities = parameters.numStartActivities;
		this.numEndActivities = parameters.numEndActivities;
		this.noIntermediateActivities = parameters.noIntermediateActivities;
		this.noOfWantedPlaces = parameters.noOfWantedPlaces;
		this.maxNoIterations = parameters.maxNoIterations;
		this.maxSourceActivitiesNumber = parameters.maxSourceActivities;
		this.maxTargetActivitiesNumber = parameters.maxTargetActivities;
		
		this.net = new PetrinetImpl("");
		this.initialMarking = new Marking();
		this.finalMarking = new Marking();
		this.sourcePlace = net.addPlace("source");
		this.sinkPlace = net.addPlace("sink");
		
		this.initialMarking.add(this.sourcePlace);
		this.finalMarking.add(this.sinkPlace);
		
		this.sourceActivitiesString = generateRandomWords(numStartActivities);
		this.targetActivitiesString = generateRandomWords(numEndActivities);
		this.intermediateActivitiesString = generateRandomWords(noIntermediateActivities);
		this.transitionCorrespondence = new HashMap<String, Transition>();
		
		this.placesMemorizer = new PlacesMemorizer();
		
		for (String sourceActivity : sourceActivitiesString) {
			Transition transition = net.addTransition(sourceActivity);
			transitionCorrespondence.put(sourceActivity, transition);
			this.net.addArc(this.sourcePlace, transitionCorrespondence.get(sourceActivity));			
		}
		for (String targetActivity : targetActivitiesString) {
			Transition transition = net.addTransition(targetActivity);
			transitionCorrespondence.put(targetActivity, transition);
			this.net.addArc(transitionCorrespondence.get(targetActivity), this.sinkPlace);
		}
		for (String intermediateActivity : intermediateActivitiesString) {
			Transition transition = net.addTransition(intermediateActivity);
			transitionCorrespondence.put(intermediateActivity, transition);
		}

	}
	
	public static List<String> generateRandomWords(int numberOfWords)
	{
	    String[] randomStrings = new String[numberOfWords];
	    Random random = new Random();
	    for(int i = 0; i < numberOfWords; i++)
	    {
	        char[] word = new char[3];
	        for(int j = 0; j < word.length; j++)
	        {
	            word[j] = (char)('a' + random.nextInt(26));
	        }
	        randomStrings[i] = new String(word).toUpperCase();
	    }
	    return Arrays.asList(randomStrings);
	}
	
	public Set<String> getSubset(List<String> possibleActivities, int maxNumber) {
		List<String> list = new LinkedList<String>(possibleActivities);
		Collections.shuffle(list);
		Random rand = new Random();
		Set<String> randomSet = new HashSet<String>(list.subList(0, Math.min(maxNumber, rand.nextInt(list.size()))));
		
		return randomSet;
	}
	
	public boolean placeAdditionTriager(Set<String> inputActivities, Set<String> outputActivities) {
		boolean ret = false;

		if (!this.placesMemorizer.checkRedundancy(inputActivities, outputActivities)) {
			
			String placeLabel = placesMemorizer.getPlaceLabel(inputActivities, outputActivities);
			Place place = net.addPlace(placeLabel);
			Set<Arc> addedArcs = new HashSet<Arc>();
			for (String inputActivity : inputActivities) {
				Transition inputTransition = this.transitionCorrespondence.get(inputActivity);
				addedArcs.add(net.addArc(inputTransition, place));
			}
			for (String outputActivity : outputActivities) {
				Transition outputTransition = this.transitionCorrespondence.get(outputActivity);
				addedArcs.add(net.addArc(place, outputTransition));
			}
			GeneratedPetrinet gnet = new GeneratedPetrinet(this.net, this.initialMarking, this.finalMarking, this.sourcePlace, this.sinkPlace, this.sourceActivitiesString, this.targetActivitiesString, this.transitionCorrespondence);
			
			ret = gnet.areAllTransitionsNonDead(this.initialMarking, this.sourcePlace);
			if (!ret) {
				for (Arc a : addedArcs) {
					this.net.removeEdge(a);
				}
				this.net.removePlace(place);
			}
			else {
				this.placesMemorizer.addPlace(inputActivities, outputActivities);
			}
		}
		
		return ret;
	}
	
	public boolean placeAdditionTriager() {		
		Set<String> subsetInput = null;
		Set<String> subsetOutput = null;
		
		while (true) {
			subsetInput = getSubset(this.intermediateActivitiesString, this.maxSourceActivitiesNumber);
			subsetOutput = getSubset(this.intermediateActivitiesString, this.maxTargetActivitiesNumber);
			
			if (subsetInput.size() > 0 && subsetOutput.size() > 0) {
				Set<String> intersection = new HashSet<String>(subsetInput);
				intersection.retainAll(subsetOutput);
				
				if (intersection.size() == 0) {
					break;
				}
			}
		}
		return placeAdditionTriager(subsetInput, subsetOutput);
	}
	
	public GeneratedPetrinet generatePetrinet() {
		int i = 0;
		int currentAddedPlaces = 0;
		
		while (i < maxNoIterations) {
			
			if (currentAddedPlaces > noOfWantedPlaces) {
				break;
			}
			boolean result = this.placeAdditionTriager();
			if (result) {
				currentAddedPlaces++;
			}
			i++;
		}
		
		GeneratedPetrinet net = new GeneratedPetrinet(this.net, this.initialMarking, this.finalMarking, this.sourcePlace, this.sinkPlace, this.sourceActivitiesString, this.targetActivitiesString, this.transitionCorrespondence);
		
		return net;
	}
	
	public static void main(String[] args) {
		PluginContextFactory factory = new PluginContextFactory();
		PluginContext context = factory.getContext();
		
		GeneratePetrinetParameters parameters = new GeneratePetrinetParameters();
		GeneratePetrinet generatePetrinet = new GeneratePetrinet(context, parameters);
		
		System.out.println("generating net");
		GeneratedPetrinet generatedNet = generatePetrinet.generatePetrinet();
		System.out.println("generated net");
		LogGeneratorParameters logParameters = new LogGeneratorParameters();
		LogGenerator logGenerator = new LogGenerator(generatedNet, logParameters);
		XLog log = logGenerator.generateXLog();
		System.out.println("generated log");
		
		int minTraceSize = 99999;
		int maxTraceSize = 0;
		
		for (XTrace tr : log) {
			if (tr.size() < minTraceSize) {
				minTraceSize = tr.size();
			}
			if (tr.size() > maxTraceSize) {
				maxTraceSize = tr.size();
			}
		}
		
		System.out.println(minTraceSize);
		System.out.println(maxTraceSize);
	}
}
