package org.processmining.glueminer.models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

public class GeneratedPetrinet {
	public Petrinet net;
	public Place sourcePlace;
	public Place sinkPlace;
	public Marking initialMarking;
	public Marking finalMarking;
	public Map<Transition, List<Place>> inputPlacesForTransitions;
	public Map<Transition, List<Place>> outputPlacesForTransitions;
	public Map<Place, List<Transition>> inputTransitionsForPlace;
	public Map<Place, List<Transition>> outputTransitionsForPlace;
	public List<Transition> transitionsAttachedToSourcePlace;
	public List<Transition> transitionsAttachedToSinkPlace;
	public List<String> startActivities;
	public List<String> endActivities;
	public Map<String, Transition> activitiesTransitionCorrespondence;
	
	public GeneratedPetrinet(Petrinet net, Marking initialMarking, Marking finalMarking, Place sourcePlace, Place sinkPlace, List<String> startActivities, List<String> endActivities, Map<String, Transition> activitiesTransitionCorrespondence) {
		this.net = net;
		this.initialMarking = initialMarking;
		this.finalMarking = finalMarking;
		this.inputPlacesForTransitions = new HashMap<Transition, List<Place>>();
		this.outputPlacesForTransitions = new HashMap<Transition, List<Place>>();
		this.inputTransitionsForPlace = new HashMap<Place, List<Transition>>();
		this.outputTransitionsForPlace = new HashMap<Place, List<Transition>>();
		this.sourcePlace = sourcePlace;
		this.sinkPlace = sinkPlace;
		this.transitionsAttachedToSourcePlace = new ArrayList<Transition>();
		this.transitionsAttachedToSinkPlace = new ArrayList<Transition>();
		this.startActivities = startActivities;
		this.endActivities = endActivities;
		this.activitiesTransitionCorrespondence = activitiesTransitionCorrespondence;
		
		this.calculateInputAndOutputPlaces();
	}
	
	public void calculateInputAndOutputPlaces() {
		for (Transition transition : net.getTransitions()) {
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inputEdges = net.getInEdges(transition);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outputEdges = net.getOutEdges(transition);
			
			this.inputPlacesForTransitions.put(transition, new ArrayList<Place>());
			this.outputPlacesForTransitions.put(transition, new ArrayList<Place>());
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inputEdges) {
				Place sourcePlace = (Place) edge.getSource();
				this.inputPlacesForTransitions.get(transition).add(sourcePlace);
			}
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : outputEdges) {
				Place targetPlace = (Place) edge.getTarget();
				this.outputPlacesForTransitions.get(transition).add(targetPlace);
			}
		}
		
		for (Place place : net.getPlaces()) {
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inputEdges = net.getInEdges(place);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outputEdges = net.getOutEdges(place);
			
			this.inputTransitionsForPlace.put(place, new ArrayList<Transition>());
			this.outputTransitionsForPlace.put(place, new ArrayList<Transition>());
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inputEdges) {
				Transition sourceTransition = (Transition) edge.getSource();
				this.inputTransitionsForPlace.get(place).add(sourceTransition);
			}
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : outputEdges) {
				Transition targetTransition = (Transition) edge.getTarget();
				this.outputTransitionsForPlace.get(place).add(targetTransition);
			}
		}
	}
	
	public Petrinet getNet() {
		return this.net;
	}
	
	public void setNet(Petrinet net) {
		this.net = net;
	}
	
	public Marking getInitialMarking() {
		return this.initialMarking;
	}
	
	public void setInitialMarking(Marking initialMarking) {
		this.initialMarking = initialMarking;
	}
	
	public List<Transition> getActivatedTransitionsAccordingToMarking(Marking marking) {
		List<Transition> activated = new ArrayList<Transition>();
		
		for (Transition t : inputPlacesForTransitions.keySet()) {
			boolean transitionActivated = true;
			
			for (Place p : inputPlacesForTransitions.get(t)) {				
				if (!marking.contains(p)) {
					transitionActivated = false;
					break;
				}
			}
			
			if (transitionActivated) {
				activated.add(t);
			}
		}
		return activated;
	}
	
	public Set<Transition> getPossiblyDeadTransitions(Marking currentMarking) {
		Set<Transition> nonDeadTransitions = new HashSet<Transition>();
		nonDeadTransitions.addAll(getActivatedTransitionsAccordingToMarking(currentMarking));
		Set<Transition> possiblyDeadTransitions = new HashSet<Transition>(activitiesTransitionCorrespondence.values());
		possiblyDeadTransitions.removeAll(nonDeadTransitions);
		
		return possiblyDeadTransitions;
	}
	
	public void addPlaceToVisitedPlaces(Map<Place, Integer> visitedPlaces, Place place) {
		if (!visitedPlaces.containsKey(place)) {
			visitedPlaces.put(place, 0);
		}
		visitedPlaces.put(place, visitedPlaces.get(place) + 1);
	}
	
	public Integer getMaxIntValueFromMap(Map<Place, Integer> visitedPlaces) {
		return Collections.max(visitedPlaces.values());
	}
	
	public Boolean areAllTransitionsNonDead(Marking currentMarking, Place sourcePlace) {
		Set<Transition> possiblyDeadTransitions = getPossiblyDeadTransitions(currentMarking);
		
		if (possiblyDeadTransitions.size() == 0) {
			return new Boolean(true);
		}
				
		Map<Place, Integer> visitedPlaces = new HashMap<Place, Integer>();
		
		/*for (Place p : currentMarking) {
			addPlaceToVisitedPlaces(visitedPlaces, p);
		}*/
		
		Set<Transition> transitionsToVisit = new HashSet<Transition>(possiblyDeadTransitions);
		
		//System.out.println("\n\n");
		while (!visitedPlaces.containsKey(sourcePlace) && !(transitionsToVisit.size() == 0)) {
			//System.out.println(transitionsToVisit.size());
			Set<Place> placesThatMustBeActive = new HashSet<Place>();
			for (Transition t : transitionsToVisit) {
				placesThatMustBeActive.addAll(inputPlacesForTransitions.get(t));
			}
			for (Place p : placesThatMustBeActive) {
				addPlaceToVisitedPlaces(visitedPlaces, p);
			}
			//System.out.println("visitedPlaces = "+visitedPlaces);
			Integer maxValue = getMaxIntValueFromMap(visitedPlaces);
			if (maxValue > 1) {
				//System.out.println("aa false");
				//System.out.println("visitedPlaces = "+visitedPlaces);
				return new Boolean(false);
			}
			//System.out.println("old transitionsToVisit = "+transitionsToVisit);
			transitionsToVisit = new HashSet<Transition>();
			for (Place p : placesThatMustBeActive) {
				transitionsToVisit.addAll(inputTransitionsForPlace.get(p));
			}
			//System.out.println("new transitionsToVisit = "+transitionsToVisit);
		}
		
		//System.out.println("true");
		
		return new Boolean(true);
	}
	
	public void activateTransition(Marking marking, Transition transition) {
		for (Place p : inputPlacesForTransitions.get(transition)) {
			marking.remove(p);
		}
		for (Place p : outputPlacesForTransitions.get(transition)) {
			marking.add(p);
		}
	}
	
	/*public List<String> activitiesToAdd(Place place, List<String> currentActivitiesToAdd, Marking currentMarking, Set<Place> visitedPlaces) {
		List<Transition> activatedTransitions = new ArrayList<Transition>(getActivatedTransitionsAccordingToMarking(currentMarking));
		visitedPlaces.add(place);
		
		for (Transition t : this.outputTransitionsForPlace.get(place)) {
			if (!activatedTransitions.contains(t)) {
				List<Place> sourcePlaces = this.inputPlacesForTransitions.get(t);
				
				for (Place p : sourcePlaces) {
					if (!currentMarking.contains(p) && !visitedPlaces.contains(p)) {
						activitiesToAdd(p, currentActivitiesToAdd, currentMarking, visitedPlaces);
						currentMarking.add(p);
					}
				}
			}
			
			activateTransition(currentMarking, t);
			currentActivitiesToAdd.add(t.getLabel());
			
			if (!currentActivitiesToAdd.contains(t.getLabel())) {
				
				//System.out.println("adding "+t.getLabel());
				//currentActivitiesToAdd.add(t.getLabel());
			}
		}
		return currentActivitiesToAdd;
	}*/
	
	public List<String> getSingleTraceActivities(double prUncompleteness, double prSkipping, double prInerting, Integer maxTraceLength) {
		List<Transition> startActivitiesTransitions = new ArrayList<Transition>();
		List<Transition> endActivitiesTransitions = new ArrayList<Transition>();
		List<Transition> intermediateActivitiesTransitions = new ArrayList<Transition>(this.activitiesTransitionCorrespondence.values());
		Marking currentMarking = new Marking(this.initialMarking);
		
		Random random = new Random();
		
		for (String startActivity : this.startActivities) {
			startActivitiesTransitions.add(activitiesTransitionCorrespondence.get(startActivity));
		}
		for (String endActivity : this.endActivities) {
			endActivitiesTransitions.add(activitiesTransitionCorrespondence.get(endActivity));
		}
		intermediateActivitiesTransitions.removeAll(startActivitiesTransitions);
		intermediateActivitiesTransitions.removeAll(endActivitiesTransitions);
		
		Transition st = startActivitiesTransitions.get(random.nextInt(startActivitiesTransitions.size()));
		Transition et = endActivitiesTransitions.get(random.nextInt(endActivitiesTransitions.size()));
		
		List<String> traceActivities = new ArrayList<String>();
		
		activateTransition(currentMarking, st);
		traceActivities.add(st.getLabel());
		
		int itNumber = 1;
		while (true) {
			List<Transition> activatedTransitions = new ArrayList<Transition>(getActivatedTransitionsAccordingToMarking(currentMarking));
			activatedTransitions.retainAll(intermediateActivitiesTransitions);
			
			if (currentMarking.size() > 0) {
				List<Place> markingPlaces = new ArrayList<Place>(currentMarking.baseSet());
				Collections.shuffle(markingPlaces);
				Place placeToConsider = markingPlaces.get(0);
				List<Transition> transitionsOnTheRight = new ArrayList<Transition>(outputTransitionsForPlace.get(placeToConsider));
				List<Transition> enabledTransitionsOnTheRight = new ArrayList<Transition>(transitionsOnTheRight);
				enabledTransitionsOnTheRight.retainAll(activatedTransitions);
				
				if (enabledTransitionsOnTheRight.size() > 0) {
					Transition ct = enabledTransitionsOnTheRight.get(0);
					activateTransition(currentMarking, ct);
					traceActivities.add(ct.getLabel());
					continue;
				}
			}
			

			Transition ct = activatedTransitions.get(random.nextInt(activatedTransitions.size()));
			activateTransition(currentMarking, ct);
			traceActivities.add(ct.getLabel());
			
			itNumber++;
			if (traceActivities.size() > maxTraceLength-1) {
				break;
			}
		}
		
		activateTransition(currentMarking, et);
		traceActivities.add(et.getLabel());

		return traceActivities;
	}
	
	public Boolean verifyFinalMarkingCouldBeReachedAnyway(Marking currentMarking, int currDepthLevel, int maxDepthLevel, Set<Set<Place>> examinedMarkings) {

		if (currentMarking.equals(finalMarking)) {
			return new Boolean(true);
		}
		else if (examinedMarkings.contains(currentMarking)) {
			return new Boolean(true);
		}
		else if (currDepthLevel > maxDepthLevel) {
			return new Boolean(true);
		}
		else {
			List<Transition> activatedTransitions = this.getActivatedTransitionsAccordingToMarking(currentMarking);
			
			if (activatedTransitions.size() > 0) {
				int i = 0;
				while (i < activatedTransitions.size()) {
					Marking nextMarking = new Marking(currentMarking);
					Set<Set<Place>> newExaminedMarkings = new HashSet<Set<Place>>(examinedMarkings);
					newExaminedMarkings.add(currentMarking.baseSet());
					
					
					if (newExaminedMarkings.equals(examinedMarkings)) {
						i++;
						continue;
					}
					for (Place p : inputPlacesForTransitions.get(activatedTransitions.get(i))) {
						nextMarking.remove(p);
					}
					for (Place p : outputPlacesForTransitions.get(activatedTransitions.get(i))) {
						nextMarking.add(p);
					}
					
					if (!nextMarking.baseSet().equals(currentMarking.baseSet())) {		
						boolean finalMarkingCouldBeReached = verifyFinalMarkingCouldBeReachedAnyway(nextMarking, currDepthLevel+1, maxDepthLevel, newExaminedMarkings);
						
						if (!finalMarkingCouldBeReached) {
							return new Boolean(false);
						}
					}
					
					i++;
				}
				return new Boolean(true);
			}
		}
		
		return new Boolean(false);
	}
}
